#include <linux/limits.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <ctype.h>
#include <dirent.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <string.h>

#ifndef PORT
#define PORT 10000
#endif

static char *password;

int check_password()
{
	// XXX: no access until we actually implement password checking
	return 0;
}

#define PASSWSZ 23
void do_setpassword(const char *pw)
{
	if (strlen(pw) != PASSWSZ) {
		fprintf(stderr, "invalid password (len %ld)\n", strlen(pw));
		return;
	}

	if (password != NULL) {
		fprintf(stderr, "password already set\n");
		return;
	}

	password = malloc(PASSWSZ + 1);
	if (password == NULL) {
		fprintf(stderr, "out of memory\n");
		return;
	}
	strcpy(password, pw);
}

void do_secret()
{
	if (!check_password()) {
		fprintf(stderr, "invalid pw\n");
		return;
	}

	// TODO: show the 'secret' file
}

#define MAXKEYS 10
char *keys[MAXKEYS];

int check_key(const char *key)
{
	for (const char *c = key; *c != '\0'; c++)
		if (!isalnum(*c))
			return 0;
	return 1;
}

void do_addkey(const char *key)
{
	int i;
	size_t s;

	for (i = 0; i < MAXKEYS; i++) {
		if (keys[i] == NULL)
			break;
	}
	if (i == MAXKEYS) {
		fprintf(stderr, "too many keys\n");;
		return;
	}
	s = strlen(key);
	if (s == 0) {
		fprintf(stderr, "key missing\n");
		return;
	}
	keys[i] = malloc(s + 1);
	if (keys[i] == NULL) {
		fprintf(stderr, "out of memory\n");
		return;
	}
	strcpy(keys[i], key);
	if (!check_key(keys[i])) {
		fprintf(stderr, "invalid key\n");
		free(keys[i]);
	}
}

#define MAXLINE 1024
void do_grep(const char *index)
{
	int i;
	char buf[MAXLINE];
	size_t s;

	i = atoi(index);
	if (i < 0 || i >= MAXKEYS) {
		fprintf(stderr, "invalid index\n");
		return;
	}

	if (keys[i] == NULL) {
		fprintf(stderr, "no such key\n");
		return;
	}

	// set -e: exit on first error
	s = snprintf(buf, MAXLINE, "set -e; /bin/grep %s public*", keys[i]);
	if (s >= MAXLINE) {
		fprintf(stderr, "key too long\n");
		return;
	}
	system(buf);
}

// NO INTENTIONAL BUGS AFTER THIS POINT
void do_listkeys()
{
	int i;

	for (i = 0; i < MAXKEYS; i++) {
		if (keys[i] != NULL) {
			printf("[%d] %s\n", i, keys[i]);
		}
	}
}

void child(int uid)
{
	char buf[MAXLINE], *n;

	printf("Secure remote grep server v2.0\n");

	for (;;) {
		if (fgets(buf, MAXLINE, stdin) == NULL)
			break;
		n = index(buf, '\n');
		if (n != NULL)
			*n = '\0';
		switch (buf[0]) {
		case '\0':
			break;
		case 'k':
			do_addkey(buf + 1);
			break;
		case 'l':
			do_listkeys();
			break;
		case 'g':
			do_grep(buf + 1);
			break;
		case 'p':
			do_setpassword(buf + 1);
			break;
		case 's':
			do_secret(buf + 1);
			break;
		}
	}
}

void sigchld(int signo)
{
	int status;
	pid_t pid = waitpid(-1, &status, WNOHANG);
	if (pid > 0) {
		printf("child %ld: ", (long)pid);
		if (WIFEXITED(status)) {
			printf("exited with status: %d\n", WEXITSTATUS(status));
		} else {
			printf("%s\n", strsignal(WTERMSIG(status)));
		}
	}
}

int main()
{
	int lstn;
	int enable;
	struct sockaddr_in lstn_addr;

	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);

	lstn = socket(AF_INET, SOCK_STREAM, 0);
	if (lstn < 0) {
		perror("socket");
		return 1;
	}
	enable = 1;
	if (setsockopt(lstn, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(enable)) < 0) {
		perror("setsockopt");
		return 1;
	}
	bzero(&lstn_addr, sizeof(lstn_addr));

	lstn_addr.sin_family = AF_INET;
	lstn_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	lstn_addr.sin_port = htons(PORT);

	if (bind(lstn, (struct sockaddr *)&lstn_addr, sizeof(lstn_addr)) < 0) {
		perror("bind");
		return 1;
	}

	if (listen(lstn, 10) < 0) {
		perror("listen");
		return 1;
	}
	printf("Listening on port %d\n", PORT);

	signal(SIGCHLD, sigchld);

	for (;;) {
		int con = accept(lstn, NULL, NULL);
		if (con < 0) {
			perror("accept");
			return 1;
		}

		switch (fork()) {
		case -1:
			perror("fork");
			return 1;
		case 0:
			printf("New connection, child %d\n", getpid());

			close(0);
			if (dup(con) < 0)
				goto error;
			close(1);
			if (dup(con) < 0)
				goto error;
			close(2);
			if (dup(con) < 0)
				goto error;
			child(1);
			exit(0);
			break;
		default:
			close(con);
			break;
		}
	}
	return 0;

error:
	fprintf(stderr, "dup() failed\n");
	exit(1);
}
